; docformat = 'rst'

;+
; This is the main interface routine for the OPeNDAP IDL bindings. 
;
; :Author: 
;    Patrick West <mrpatrickwest@gmail.com> UCAR/HAO
;
;    Dan Holloway <d.holloway@opendap.org>
;
;    Based on IDL-DODS Written by Daniel J. Carr - Research Systems, Inc.
;
; :Copyright: 
;    Copyright (c) 2006 OPeNDAP, Inc.
;
;    This is free software; you can redistribute it and/or
;    modify it under the terms of the GNU Lesser General Public
;    License as published by the Free Software Foundation; either
;    version 2.1 of the License, or (at your option) any later version.
; 
;    This library is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;    Lesser General Public License for more details.
; 
;    You should have received a copy of the GNU Lesser General Public
;    License along with this library; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;    You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
;- 

;+
; Main interface routine for IDL-OPENDAP link.
;
; opendap_get sets up the environment, and then calls opendap_get_das or 
; opendap_get_data to establish a connection and return the appropriate 
; OPeNDAP data.
;
; The DAP data ("opendap_data" parameter) is returned as an IDL structure 
; variable. The IDL structure will have one tag for each DAP variable that is 
; returned, and the tag name is the same as the DAP name for that variable. 
; For simple DAP variables, the value of the corresponding IDL structure tag 
; will be an IDL scalar value. Other DAP variable types will be returned as 
; nested structures::
;
;    DAP data type         Returned IDL structure
;    --------------------- ------------------------------------------------
; 
;    ARRAY                   { ARRAY: data, 
;                              START_STRIDE_STOP: fltarr(3, N),
;                              DIMENSION_NAMES: strarr(N) }
;
;    GRID                    { GRID: data,
;                              dimension_name_0: dimension_data_0,
;                              ...
;                              dimension_name_n: dimension_data_n }
;
;                            NOTE: the grid dimension names are optional -
;                            not all grids will have them.
;
;    STRUCTURE               { var_name_0: var_data_0,
;                              ...
;                              var_name_n: var_data_n }
;
;    SEQUENCE                { record_name_0: record_data_0,
;                              ...
;                              record_name_n: record_data_n }
;
;                            NOTE: each record_data field corresponds to a
;                            sequence "column".
;
; Tag names in capital letters are literal. Other tag names are derived from 
; the DODS data.
;
; :Returns:
;    Status (1=ok, 0=fail)
;
; :Params:
;    url : in, required, type=string
;       URL location to connect to.
;    opendap_data : out, optional, type=structure
;       Returns the requested DODS data as an IDL structure.
;
; :Keywords:
;    ce : in, optional, type=string
;       Data constraint expression.
;    deflate :in, optional, type=boolean, default=1
;       ACCEPT_DEFLATE flag  (1=yes (default), 0=no).
;    mode : in, optional, type=string
;       Flag for the type of data to retrieve 
;       (0 or 'DATA' for the data (default), 1 or 'DAS' for the DAS) 
;    seq_blocksize : in, optional, type=long, default=4096
;       How many elements should be allocated at once for sequence data;
;       default value is 4096
;    debug : in, optional, type=boolean
;       set to print out debug information
;    indicate_type : in, optional, type=boolean
;       set to have each structure in the return value indicate it's type 
;       (i.e. "GRID", "GROUP", etc.)
; 
; :Examples:
;    Using OPENDAP_GET to retrieve all the data at an URL is quite simple::
;
;       IDL> url = 'http://test.opendap.org/opendap-3.4/nph-dods/data/nc/fnoc1.nc'
;       IDL> stat = opendap_get(url, data)
;       IDL> print, stat
;                  1
;       IDL> help, data,  /structures
;       ** Structure <824e1a4>, 5 tags, length=23172, data length=23172, refs=1:
;          U               STRUCT    -> <Anonymous> Array[1]
;          V               STRUCT    -> <Anonymous> Array[1]
;          LAT             STRUCT    -> <Anonymous> Array[1]
;          LON             STRUCT    -> <Anonymous> Array[1]
;          TIME            STRUCT    -> <Anonymous> Array[1]
;       IDL> print, tag_names(data)
;       U V LAT LON TIME
;       IDL> help, data.u, /structures
;       ** Structure <8252124>, 2 tags, length=11460, data length=11460, refs=2:
;          U               INT       Array[21, 17, 16]
;          DIMENSION_NAMES STRING    Array[3]
;       IDL> print, data.u.dimension_names
;       LON LAT TIME_A
;
;    Single variables or even slices of a variable can also be retrieved::
;
;       IDL> help, data, /structures
;       ** Structure <1f29584>, 1 tags, length=11460, data length=11460, refs=1:
;          U               STRUCT    -> <Anonymous> Array[1]
;       IDL> help, data.u, /structures
;       ** Structure <1f29974>, 2 tags, length=11460, data length=11460, refs=2:
;          U               INT       Array[21, 17, 16]
;          DIMENSION_NAMES STRING    Array[3]
;
;    or::
; 
;       IDL> stat = opendap_get(url, data, ce='u[0][0:1:16][0:1:15]')
;       IDL> help, data, /structures
;       ** Structure <1f29af4>, 1 tags, length=568, data length=568, refs=1:
;          U               STRUCT    -> <Anonymous> Array[1]
;       IDL> help, data.u, /structures
;       ** Structure <1f29ef4>, 2 tags, length=568, data length=568, refs=2:
;          U               INT       Array[16, 17]
;          DIMENSION_NAMES STRING    Array[2]
;
;    Attributes can be optained as well::
;
;       IDL> stat = opendap_get(url, data, mode='das')               
;       IDL> help, data, /structures
;       ** Structure <207b604>, 7 tags, length=212, data length=209, refs=1:
;          U               STRUCT    -> <Anonymous> Array[1]
;          V               STRUCT    -> <Anonymous> Array[1]
;          LAT             STRUCT    -> <Anonymous> Array[1]
;          LON             STRUCT    -> <Anonymous> Array[1]
;          TIME            STRUCT    -> <Anonymous> Array[1]
;          NC_GLOBAL       STRUCT    -> <Anonymous> Array[1]
;          DODS_EXTRA      STRUCT    -> <Anonymous> Array[1]
;       IDL> help, data.u, /structures
;       ** Structure <1f2b884>, 8 tags, length=80, data length=77, refs=2:
;          UNITS           STRING    'meter per second'
;          LONG_NAME       STRING    'Vector wind eastward component'
;          MISSING_VALUE   STRING    '-32767'
;          SCALE_FACTOR    STRING    '0.005'
;          DODS_NAME       STRING    'UWind'
;          B               BYTE       128
;          I               LONG             32000
;          WOA01           STRING    'http://localhost/junk'
;-
FUNCTION OPENDAP_GET, url, opendap_data, CE=opendap_ce, $
                      DEFLATE=deflate, MODE=mode, $
                      SEQ_BLOCKSIZE=seq_blocksize, $
                      DEBUG=debug, INDICATE_TYPE=indicateType
  COMPILE_OPT strictarr
  
  ; If this is the first time run in this IDL session,
  ; then define system variables and link external routines.
  DEFSYSV, '!opendap_null_c', EXISTS=opendap_defined  
  IF (opendap_defined EQ 0) THEN BEGIN
    @opendap_setup
  ENDIF

  IF (N_ELEMENTS(url) LE 0UL) THEN BEGIN
    PRINT, 'url passed is empty'
    RETURN, 0ul
  ENDIF
  
  IF (STRCOMPRESS(url) EQ '') THEN BEGIN
    PRINT, 'url passed is empty'
    RETURN, 0ul
  ENDIF

  ; set defaults
  
  IF (N_ELEMENTS(opendap_ce) LE 0L) THEN opendap_ce = ''
  IF (N_ELEMENTS(deflate) LE 0UL) THEN deflate = 0UL
  IF (N_ELEMENTS(debug) LE 0UL) THEN debug = 0UL
  IF (N_ELEMENTS(seq_blocksize) LE 0UL) THEN seq_blocksize = 4096UL      
  IF (N_ELEMENTS(mode) EQ 0UL) THEN mode = 'DATA'   
     
  CASE STRUPCASE(mode) OF
    'DATA': mode = 0UL
    'DDS': mode = 1UL
    'DAS': mode = 2UL
    'FULLDATA': mode = 3UL
    'FULLDDS': mode = 4UL
    'FULLDAS': mode = 5UL
    ELSE: BEGIN
        PRINT, 'Unrecognized mode, MODE=', STRING(mode)
        PRINT, 'Options are: DDS, DAS, DATA, FULLDDS, FULLDAS, and FULLDATA'
        RETURN, 0UL          
      END
  ENDCASE

  stat = 0UL

  IF (mode EQ 0UL OR mode EQ 1UL OR mode EQ 3UL OR mode EQ 4UL) THEN BEGIN
    stat = OPENDAP_GET_DATA(url, opendap_data, CE=opendap_ce, $
                            DEFLATE=deflate, $
                            SEQ_BLOCKSIZE=seq_blocksize, $
                            MODE=mode, debug, INDICATE_TYPE=indicateType)
  ENDIF ELSE BEGIN 
    IF (mode EQ 2UL OR mode EQ 5UL) THEN BEGIN
      stat = OPENDAP_GET_DAS(url, opendap_data, CE=opendap_ce, $
                             DEFLATE=deflate, MODE=mode, debug, $
                             INDICATE_TYPE=indicateType)
    ENDIF
  ENDELSE

  RETURN, stat
END

