; docformat = 'rst'

;+
; IDL-OPENDAP interface.
;
; :Private:
;
; :Author: 
;    Patrick West <mrpatrickwest@gmail.com> UCAR/HAO
;
;    Based on IDL-DODS Written by Daniel J. Carr - Research Systems, Inc.
;
; :Copyright:
;    Copyright (c) 2006 OPeNDAP, Inc.
;
;    This is free software; you can redistribute it and/or
;    modify it under the terms of the GNU Lesser General Public
;    License as published by the Free Software Foundation; either
;    version 2.1 of the License, or (at your option) any later version.
; 
;    This library is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;    Lesser General Public License for more details.
; 
;    You should have received a copy of the GNU Lesser General Public
;    License along with this library; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;    You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
;-

;+
; Package up the content of a node as a structure, depending on the type of
; request (depending on get_data).
;
; :Params:
;    name : in, required, type=string
;       name of the content
;    arr : in, required, type=array
;       content
;    get_data : in, required, type=boolean
;       1 to return actual data, 0 to return size/type information
;
; :Keywords:
;    d_names : in, optional, type=strarr
;       dimension names
;    indicate_type : in, optional, type=boolean
;       set to have each structure in the return value indicate it's type 
;       (i.e. "GRID", "GROUP", etc.)
;-
FUNCTION opendap_package_content, name, arr, get_data, $
                                  D_NAMES=d_names, INDICATE_TYPE=indicateType
  COMPILE_OPT strictarr
  
  arr_dims = SIZE(arr, /N_DIMENSIONS)
  d_dims = N_ELEMENTS(d_names)
  
  IF (get_data) THEN BEGIN
    ; Build IDL structure to hold the data
    opendap_data = CREATE_STRUCT(STRING(name), $
                                 TEMPORARY(arr))
  ENDIF ELSE BEGIN
    ; Build IDL structure to hold the type information
    opendap_data = CREATE_STRUCT('n_dimensions', $
                                 arr[0], $
                                 'dimensions', $
                                 arr[1:arr[0]], $
                                 'type', $
                                 arr[arr[0] + 1], $
                                 'n_elements', $
                                 PRODUCT(arr[1:arr[0]], /PRESERVE_TYPE))  
  ENDELSE  
  
  ; tack on dimensions_names in either case, if d_names is present
  IF (d_dims gt 0L) THEN BEGIN
    opendap_data = CREATE_STRUCT(opendap_data, $
                                'dimension_names', $
                                REVERSE(d_names[0L:d_dims - 1L]))
  ENDIF
               
  IF (KEYWORD_SET(indicateType)) THEN BEGIN
    opendap_data = CREATE_STRUCT('_TYPE', 'GRID:MAP', opendap_data)
  ENDIF
                                    
  RETURN, opendap_data                               
END

                                 
;+
; Handle retrieving content for a given node.
;
; :Params:
;    opendap_data : in, out, required, type=structure
;       returned data for node
;    node : in, required, type=long
;       node ID
;    content : in, required, type=long
;       content pointer
;    node_type : in, required, type=long
;       node type code as defined by !opendap_NODE_xxxx system variables
;    has_content : in, required, type=int
;       set to 1 
;    get_data : in, required, type=int
;       set to 1 to get_data, 0 to not get any data
;    debug : in, required, type=int
;       set to 1 to debug
;
; :Keywords:
;    is_empty : out, optional, type=boolean
;       set to 1 if returning empty sequence
;    name : in, optional, type=string
;       name of the node, used if the node can't find its own name
;    indicate_type : in, optional, type=boolean
;       set to have each structure in the return value indicate it's type 
;       (i.e. "GRID", "GROUP", etc.)
;-
FUNCTION opendap_handle_content, opendap_data, node, content, node_type, $
                                 has_content, get_data, debug, $
                                 IS_EMPTY=isEmpty, NAME=nodeName, $
                                 INDICATE_TYPE=indicateType
  COMPILE_OPT strictarr
  
  _nodeName = N_ELEMENTS(nodeName) EQ 0L ? 'placeholder' : nodeName
  
  ; Based on the type of the variable, call the appropriate function
  ; from idl_dods.cc to get the data
  IF (debug gt 0UL) THEN PRINT, 'node_type = ', node_type
  
  CASE node_type OF 
    !opendap_NODE_Byte: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'byte'
        status = OPENDAP_GET_BYTE(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          opendap_data = (REFORM(TEMPORARY(val)))[0]
        ENDIF ELSE RETURN, 0UL
      END
        
    !opendap_NODE_Int16: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'int16'
        status = OPENDAP_GET_INT16(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          opendap_data = (REFORM(TEMPORARY(val)))[0]
        ENDIF ELSE RETURN, 0UL
      END
      
    !opendap_NODE_UInt16: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'uint16'
        status = OPENDAP_GET_UINT16(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          opendap_data = (REFORM(TEMPORARY(val)))[0]
        ENDIF ELSE RETURN, 0UL 
      END
      
    !opendap_NODE_Int32: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'int32'
        status = OPENDAP_GET_INT32(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          opendap_data = (REFORM(TEMPORARY(val)))[0]
        ENDIF ELSE RETURN, 0UL
      END
      
    !opendap_NODE_UInt32: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'uint32'
        status = OPENDAP_GET_UINT32(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          opendap_data = (REFORM(TEMPORARY(val)))[0]
        ENDIF ELSE RETURN, 0UL
      END
        
    !opendap_NODE_Float32: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'float32'
        status = OPENDAP_GET_FLOAT32(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          opendap_data = (REFORM(TEMPORARY(val)))[0]
        ENDIF ELSE RETURN, 0UL
      END
      
    !opendap_NODE_Float64: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'float64'
        status = OPENDAP_GET_FLOAT64(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          opendap_data = (REFORM(TEMPORARY(val)))[0]
        ENDIF ELSE RETURN, 0UL
      END
      
    !opendap_NODE_String: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'string'
        status = OPENDAP_GET_STRING(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          ; The STRTRIM shouldn't have any effect here, however
          ; if it's removed, the sequence code returns the last value
          ; for every value of the array
          opendap_data = STRTRIM(TEMPORARY(val))
        ENDIF ELSE RETURN, 0UL
      END
        
    !opendap_NODE_URL: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'url'
        status = OPENDAP_GET_STRING(content, val, get_data)
        IF ((status NE 0UL) AND (N_ELEMENTS(val) GT 0UL)) THEN BEGIN
          ; The STRTRIM shouldn't have any effect here, however
          ; if it's removed, the sequence code returns the last value
          ; for every value of the array
          opendap_data = STRTRIM(TEMPORARY(val))
        ENDIF ELSE RETURN, 0UL
      END
        
    !opendap_NODE_Array: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'array'
        d_names = BYTARR(65, 12)
        status = OPENDAP_GET_ARRAY(content, arr, arr_name, d_names, get_data)
        IF (debug GT 0UL) THEN PRINT, 'n_elements(arr) = ', n_elements(arr)  
        IF ((status NE 0UL) AND (N_ELEMENTS(arr) GT 0UL)) THEN BEGIN
          arr_size = SIZE(arr)
          arr_dims = arr_size[0]

          ; Figure out how many dimensions are actually used
          d_dims = 0
          FOR j = 0, 11 DO BEGIN
            IF (d_names[0, j] NE 0) THEN d_dims = d_dims + 1
          ENDFOR

          ; And create an array of the appropriate size to hold them
          str_names = STRARR(d_dims)
          FOR i = 0L, d_dims - 1 DO BEGIN
            temp_name = REFORM( d_names[*,i] )
            opendap_fix_name, temp_name
            str_names[i] = temp_name
          ENDFOR

          OPENDAP_FIX_NAME, arr_name
          
          opendap_data = OPENDAP_PACKAGE_CONTENT(STRING(arr_name), $
                                                 TEMPORARY(arr), $
                                                 get_data, $
                                                 D_NAMES=str_names)
          IF (KEYWORD_SET(indicateType)) THEN BEGIN
            opendap_data = CREATE_STRUCT('_TYPE', 'ARRAY', opendap_data)
          ENDIF
        ENDIF ELSE RETURN, 0UL
      END
      
    !opendap_NODE_Grid: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'grid'
        
        ; First lets get the values of the grid, the dimension names,
        ; the name of the array, etc...
        IF (debug GT 0UL) THEN PRINT, 'getting array'
        OPENDAP_GET_GRID_VALUES, content, vcontent
        IF (debug GT 0UL) THEN PRINT, 'handling array'
        status = OPENDAP_HANDLE_CONTENT(vals, node, vcontent, $
                                        !opendap_NODE_Array, $
                                        has_content, $
                                        get_data, debug, $
                                        INDICATE_TYPE=indicateType)
        IF (debug GT 0UL) THEN PRINT, 'done getting array = ', status

        IF ((status NE 0UL)) THEN BEGIN
          g_names = tag_names( vals )
          g_name = g_names[0]
          IF (debug GT 0UL) THEN PRINT, 'g_name = ', g_name
          IF (debug GT 0UL) THEN PRINT, 'getting grid'
          m_names = BYTARR(64, 12) ; map dimension names
          
          ; Now lets get the maps
          status = OPENDAP_GET_GRID(content, m1, m2, m3, m4, m5, m6, $
                                    m7, m8, m9, m10, m11, m12, $
                                    m_names, get_data )
          IF (debug GT 0UL) THEN PRINT, '  done getting grid: ', status
          IF (status NE 0UL) THEN BEGIN
            str_names = STRARR(12)
            n_names = N_ELEMENTS(m_names)
            FOR i_names = 0, n_names - 1 DO BEGIN
              temp_name = m_names[i_names]
              OPENDAP_FIX_NAME, temp_name
              str_names[i_names] = temp_name
            ENDFOR
            d_names = TEMPORARY(str_names)
            IF (debug GT 0UL) THEN BEGIN
              PRINT, '  n_names = ', n_names
              FOR i_names = 0, n_names - 1 DO BEGIN
                  PRINT, '  d_names[', i_names,'] = ', d_names[i_names]
              ENDFOR
              IF (N_ELEMENTS(m1) GT 0UL) THEN PRINT, '  m1 has data'
              IF (N_ELEMENTS(m2) GT 0UL) THEN PRINT, '  m2 has data'
              IF (N_ELEMENTS(m3) GT 0UL) THEN PRINT, '  m3 has data'
              IF (N_ELEMENTS(m4) GT 0UL) THEN PRINT, '  m4 has data'
              IF (N_ELEMENTS(m5) GT 0UL) THEN PRINT, '  m5 has data'
              IF (N_ELEMENTS(m6) GT 0UL) THEN PRINT, '  m6 has data'
              IF (N_ELEMENTS(m7) GT 0UL) THEN PRINT, '  m7 has data'
              IF (N_ELEMENTS(m8) GT 0UL) THEN PRINT, '  m8 has data'
              IF (N_ELEMENTS(m9) GT 0UL) THEN PRINT, '  m9 has data'
              IF (N_ELEMENTS(m10) GT 0UL) THEN PRINT, '  m10 has data'
              IF (N_ELEMENTS(m11) GT 0UL) THEN PRINT, '  m11 has data'
              IF (N_ELEMENTS(m12) GT 0UL) THEN PRINT, '  m12 has data'
            ENDIF

            opendap_data = CREATE_STRUCT(_nodeName, REFORM(TEMPORARY(vals)))

            ; Build IDL structure to hold the components.
            IF (N_ELEMENTS(m1) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[0])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[0] = d_names[0] + '_1'
              ENDIF       
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[0], $
                                                 TEMPORARY(m1), $
                                                 get_data, $
                                                 D_NAMES=d_names[0], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[0], map_data)           
            ENDIF
                
            IF (N_ELEMENTS(m2) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[1])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[1] = d_names[1] + '_1'
              ENDIF
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[1], $
                                                 TEMPORARY(m2), $
                                                 get_data, $
                                                 D_NAMES=d_names[1], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[1], map_data)            
            ENDIF
                
            IF (N_ELEMENTS(m3) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[2])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[2] = d_names[2] + '_1'
              ENDIF

              map_data = OPENDAP_PACKAGE_CONTENT(d_names[2], $
                                                 TEMPORARY(m3), $
                                                 get_data, $
                                                 D_NAMES=d_names[2], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[2], map_data)
            ENDIF
                                    
            IF (N_ELEMENTS(m4) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[3])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[3] = d_names[3] + '_1'
              ENDIF
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[3], $
                                                 TEMPORARY(m4), $
                                                 get_data, $
                                                 D_NAMES=d_names[3], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[3], map_data)
            ENDIF
                
            IF (N_ELEMENTS(m5) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[4])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[4] = d_names[4] + '_1'
              ENDIF
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[4], $
                                                 TEMPORARY(m5), $
                                                 get_data, $
                                                 D_NAMES=d_names[4], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[4], map_data)
            ENDIF
                
            IF (N_ELEMENTS(m6) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[5])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[5] = d_names[5] + '_1'
              ENDIF
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[5], $
                                                 TEMPORARY(m6), $
                                                 get_data, $
                                                 D_NAMES=d_names[5], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[5], map_data)
            ENDIF
                
            IF (N_ELEMENTS(m7) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[6])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[6] = d_names[6] + '_1'
              ENDIF
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[6], $
                                                 TEMPORARY(m7), $
                                                 get_data, $
                                                 D_NAMES=d_names[6], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[6], map_data)
            ENDIF
                
            IF (N_ELEMENTS(m8) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[7])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[7] = d_names[7] + '_1'
              ENDIF
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[7], $
                                                 TEMPORARY(m8), $
                                                 get_data, $
                                                 D_NAMES=d_names[7], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[7], map_data)
            ENDIF
                
            IF (N_ELEMENTS(m9) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[8])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[8] = d_names[8] + '_1'
              ENDIF
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[8], $
                                                 TEMPORARY(m9), $
                                                 get_data, $
                                                 D_NAMES=d_names[8], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[8], map_data)
            ENDIF
            
            IF (N_ELEMENTS(m10) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[9])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[9] = d_names[9] + '_1'
              ENDIF

              map_data = OPENDAP_PACKAGE_CONTENT(d_names[9], $
                                                 TEMPORARY(m10), $
                                                 get_data, $
                                                 D_NAMES=d_names[9], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[9], map_data)
            ENDIF
                
            IF (N_ELEMENTS(m11) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[10])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[10] = d_names[10] + '_1'
              ENDIF
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[10], $
                                                 TEMPORARY(m11), $
                                                 get_data, $
                                                 D_NAMES=d_names[10], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[10], map_data)
            ENDIF
            
            IF (N_ELEMENTS(m12) GT 0UL) THEN BEGIN
              ind = WHERE(TAG_NAMES(opendap_data) EQ d_names[11])
              IF (ind[0] GE 0UL) THEN BEGIN
                d_names[11] = d_names[11] + '_1'
              ENDIF
              
              map_data = OPENDAP_PACKAGE_CONTENT(d_names[11], $
                                                 TEMPORARY(m12), $
                                                 get_data, $
                                                 D_NAMES=d_names[11], $
                                                 INDICATE_TYPE=indicateType)
              opendap_data = CREATE_STRUCT(opendap_data, d_names[11], map_data)
            ENDIF
            
            IF (KEYWORD_SET(indicateType)) THEN BEGIN
              opendap_data = CREATE_STRUCT('_TYPE', 'GRID', opendap_data)
            ENDIF
          ENDIF ELSE RETURN, 0UL
        ENDIF ELSE RETURN, 0UL
      END
        
    !opendap_NODE_Sequence: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'sequence'
        
        ; get information on the sequence, numbers of rows and number of
        ; columns.
        OPENDAP_GET_SEQUENCE_INFO, content, n_rows, n_fields
        IF (debug GT 0UL) THEN BEGIN
          PRINT, 'n_rows = ', n_rows
          PRINT, 'n_fields = ', n_fields
        ENDIF
        IF (n_rows EQ 0UL) THEN BEGIN
          n_rows = 1UL
        ENDIF
        FOR i_fields = 0UL, n_fields - 1 DO BEGIN
          stat = OPENDAP_GET_SEQUENCE_FIELD(node, content, i_fields, $
                                            field_name, field_type, $
                                            field_node, field_content)
          opendap_fix_name, field_name
          IF (stat NE 1UL) THEN RETURN, 0UL
          IF (debug GT 0UL) THEN BEGIN
            PRINT, '  field[', i_fields, '] name: ', field_name, $
                   ' type: ', field_type
          ENDIF
          
          FOR i_rows = 0UL, n_rows - 1 DO BEGIN
            stat = OPENDAP_GET_SEQUENCE_ROW(field_node, $
                                            field_content, $
                                            field_type, $
                                            i_rows, $
                                            row_value )

            IF (stat NE 1UL) THEN BEGIN
              PRINT, 'Sequence field ', i_fields, $
                     ': failed to retrieve row ', i_rows
              RETURN,0UL
            ENDIF
            
            IF (field_type EQ !opendap_NODE_Structure) THEN BEGIN
              stat = OPENDAP_HANDLE_NODE(field_data, field_node, $
                                         get_data, debug)
            ENDIF ELSE BEGIN
              stat = OPENDAP_HANDLE_CONTENT(field_data, field_node, $
                                            row_value, field_type, $
                                            has_content, $
                                            get_data, debug)
            ENDELSE

            IF (stat NE 1UL) THEN BEGIN
              PRINT, 'Sequence field ',i_fields, $
                     ': failed to retrieve row ', i_rows, ' content'
              RETURN, 0UL
            ENDIF
            
            IF (i_rows EQ 0UL) THEN BEGIN
              vals = [field_data]
            ENDIF ELSE BEGIN
              vals = [vals, field_data]
            ENDELSE
          ENDFOR

          IF (i_fields EQ 0UL) THEN BEGIN
            opendap_data = CREATE_STRUCT(field_name, $
                                         REFORM(TEMPORARY(vals)))
          ENDIF ELSE BEGIN
            opendap_data = CREATE_STRUCT(opendap_data, field_name, $
                                         REFORM(TEMPORARY(vals)))
          ENDELSE
        ENDFOR
        
          IF (KEYWORD_SET(indicateType)) THEN BEGIN
          opendap_data = CREATE_STRUCT('_TYPE', 'ARRAY', opendap_data)
        ENDIF        
      END
      
    !opendap_NODE_Structure: BEGIN
        IF (debug GT 0UL) THEN PRINT, 'structure'
        
        ; To get here there are no attributes and no subnodes, so this
        ; is an empty structure
        OPENDAP_GET_STRUCT_INFO, content, ctMembers
        IF (debug GT 0UL) THEN BEGIN
          PRINT, '  structure contains ', ctMembers, ' members'
        ENDIF
        IF (ctMembers EQ 0UL) THEN BEGIN
          opendap_data = CREATE_STRUCT('EMPTY', '')
          isEmpty = 1B
        ENDIF ELSE BEGIN 
          ; This code should not be used, so just returning
          PRINT, '  should not be getting here'
          RETURN, 0UL
          
          FOR i_ct = 0, ctMembers - 1 DO BEGIN
            IF (debug GT 0UL) THEN BEGIN
                PRINT, '  retrieving member ', i_ct
            ENDIF
            stat = OPENDAP_GET_STRUCT_MEMBER(content, i_ct, $
                                             m_type, m_content)
            IF (stat NE 1UL) THEN BEGIN
              PRINT, 'Failed to retrieve structure member ',i_ct
              RETURN, 0UL
            ENDIF
            IF (debug GT 0UL) THEN BEGIN
              PRINT, '  handling member ', i_ct
            ENDIF
            stat = OPENDAP_HANDLE_CONTENT(m_data, node, m_content, $
                                          m_type, has_content, $
                                          get_data, debug)
            IF( stat NE 1UL ) THEN BEGIN
              PRINT, 'Failed to retrieve structure member ', i_ct, $
                     ' content'
              RETURN, 0UL
            ENDIF
          ENDFOR
        ENDELSE
      END
    
    !opendap_NODE_root: BEGIN
        ; To get here there is nothing in the root node, no attributes
        ; and no content and no subnodes.
        opendap_data = CREATE_STRUCT('EMPTY', '')
      END
      
    ELSE: BEGIN
        PRINT, 'Unknown content type', node_type, ' in opendap_handle_content'
        RETURN, 0UL
      END
  ENDCASE

  RETURN, 1UL
END

